@@ -6,13 +6,13 @@
 
 __author__ = """luphord"""
 __email__ = """luphord@protonmail.com"""
-__version__ = """0.4.0"""
+__version__ = """0.5.0"""
 
 
 from argparse import ArgumentParser, Namespace
 from abc import ABC, abstractmethod
 from itertools import groupby
-from typing import Final, Union, Mapping, Tuple, Set, Iterable
+from typing import Final, Optional, Union, Mapping, Tuple, Set, Iterable, Callable
 from numbers import Real
 import numpy as np
 from dataclasses import dataclass
@@ -199,8 +199,15 @@ class DateIndex:
         return DateIndex(idx)
 
 
+class TermStructuresModel(ABC):
+    @abstractmethod
+    def linear_rate(self, frequency: str) -> np.ndarray:
+        pass
+
+
 class Model:
     dategrid: Final[np.ndarray]
+    simulated_rates: Final[Mapping[str, TermStructuresModel]]
     simulated_fx: Final[Mapping[Tuple[str, str], np.ndarray]]
     simulated_stocks: Final[Mapping[str, np.ndarray]]
     numeraire: Final[np.ndarray]
@@ -212,6 +219,7 @@ class Model:
     def __init__(
         self,
         dategrid: np.ndarray,
+        simulated_rates: Mapping[str, TermStructuresModel],
         simulated_fx: Mapping[Tuple[str, str], np.ndarray],
         simulated_stocks: Mapping[str, np.ndarray],
         numeraire: np.ndarray,
@@ -228,6 +236,7 @@ class Model:
         self.shape = (self.nsim, self.ndates)
         assert numeraire.shape == self.shape
         self.numeraire_currency = numeraire_currency
+        self.simulated_rates = simulated_rates
         for fxkey, val in simulated_fx.items():
             assert (
                 val.dtype == np.float
@@ -364,6 +373,64 @@ class ObservableFloat(ABC):
     def simulate(self, model: Model) -> np.ndarray:
         pass
 
+    def __add__(self, other: Union["ObservableFloat", float]) -> "ObservableFloat":
+        if isinstance(other, ObservableFloat):
+            return Sum(self, other)
+        elif isinstance(other, Real):
+            return Sum(self, KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __radd__(self, other: float) -> "ObservableFloat":
+        return Sum(KonstFloat(other), self)
+
+    def __neg__(self) -> "ObservableFloat":
+        return Minus(self)
+
+    def __sub__(self, other: Union["ObservableFloat", float]) -> "ObservableFloat":
+        if isinstance(other, ObservableFloat):
+            return Sum(self, -other)
+        elif isinstance(other, Real):
+            return Sum(self, -KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __rsub__(self, other: float) -> "ObservableFloat":
+        return Sum(KonstFloat(other), -self)
+
+    def __mul__(self, other: Union["ObservableFloat", float]) -> "ObservableFloat":
+        if isinstance(other, ObservableFloat):
+            return Product(self, other)
+        elif isinstance(other, Real):
+            return Product(self, KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __rmul__(self, other: float) -> "ObservableFloat":
+        return Product(KonstFloat(other), self)
+
+    def __truediv__(self, other: Union["ObservableFloat", float]) -> "ObservableFloat":
+        if isinstance(other, ObservableFloat):
+            return Quotient(self, other)
+        elif isinstance(other, Real):
+            return Quotient(self, KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __rtruediv__(self, other: float) -> "ObservableFloat":
+        return Quotient(KonstFloat(other), self)
+
+    def __pow__(self, other: Union["ObservableFloat", float]) -> "ObservableFloat":
+        if isinstance(other, ObservableFloat):
+            return Power(self, other)
+        elif isinstance(other, Real):
+            return Power(self, KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __rpow__(self, other: float) -> "ObservableFloat":
+        return Power(KonstFloat(other), self)
+
     def __ge__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
         if isinstance(other, ObservableFloat):
             return GreaterOrEqualThan(self, other)
@@ -387,6 +454,97 @@ class ObservableFloat(ABC):
         return ~(self >= other)
 
 
+@dataclass
+class Sum(ObservableFloat):
+    """Equal to the sum of two observables"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) + self.observable2.simulate(model)
+
+    def __str__(self) -> str:
+        return f"({self.observable1}) + ({self.observable2})"
+
+
+@dataclass
+class Minus(ObservableFloat):
+    """Negative value of observable"""
+
+    observable: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return -self.observable.simulate(model)
+
+    def __str__(self) -> str:
+        return f"-{self.observable}"
+
+
+@dataclass
+class Product(ObservableFloat):
+    """Equal to the product (multiplication) of two observables"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) * self.observable2.simulate(model)
+
+    def __str__(self) -> str:
+        return f"({self.observable1}) * ({self.observable2})"
+
+
+@dataclass
+class Quotient(ObservableFloat):
+    """Equal to the quotient (division) of two observables"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) / self.observable2.simulate(model)
+
+    def __str__(self) -> str:
+        return f"({self.observable1}) / ({self.observable2})"
+
+
+@dataclass
+class Power(ObservableFloat):
+    """Equal to observable1 to the power of observable2"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) ** self.observable2.simulate(model)
+
+    def __str__(self) -> str:
+        return f"({self.observable1}) ** ({self.observable2})"
+
+
+@dataclass
+class FixedAfter(ObservableFloat):
+    """Equal to observable, but remains constant as soon as
+    fixing_condition becomes true (seen from the beginning,
+    not from acquisition date)"""
+
+    fixing_condition: "ObservableBool"
+    observable: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        underlying = self.observable.simulate(model).copy()
+        fixing_times = self.fixing_condition.simulate(model)
+        fixing_idx = model.eval_date_index.next_after(fixing_times)
+        for i in range(1, underlying.shape[1]):
+            fixing_mask = (i > fixing_idx.index) & (fixing_idx.index >= 0)
+            underlying[fixing_mask, i] = underlying[fixing_mask, i - 1]
+        return underlying
+
+    def __str__(self) -> str:
+        return f"FixedAfter({self.fixing_condition}, {self.observable})"
+
+
 @dataclass
 class Stock(ObservableFloat):
     """Value of the stock identified by identifier"""
@@ -415,6 +573,21 @@ class FX(ObservableFloat):
         return f"FX({self.base_currency}/{self.counter_currency})"
 
 
+@dataclass
+class LinearRate(ObservableFloat):
+    """Value of the linear rate (e.g. a LIBOR) with payment frequency
+    in currency"""
+
+    currency: str
+    frequency: str
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return model.simulated_rates[self.currency].linear_rate(self.frequency)
+
+    def __str__(self) -> str:
+        return f"{self.currency}{self.frequency})"
+
+
 @dataclass
 class KonstFloat(ObservableFloat):
     """Always equal to constant"""
@@ -812,7 +985,7 @@ class StochasticProcess(ABC):
         actual_stddev = paths.std(axis=0)
         assert (
             target_stddev[actual_stddev == 0] == 0
-        ), "Cannot scale actual standard deviation of zero to any other value than zero"
+        ).all(), "Cannot scale actual stddev of zero to any other value than zero"
         actual_stddev[actual_stddev == 0] = 1  # prevent division by zero
         return (paths - actual_mean) / actual_stddev * target_stddev + target_mean
 
@@ -825,8 +998,10 @@ class StochasticProcess(ABC):
 class BrownianMotion(StochasticProcess):
     """Brownian Motion (Wiener Process) with optional drift."""
 
-    def __init__(self, mu: float = 0.0, sigma: float = 1.0):
-        self.mu = mu
+    def __init__(
+        self, mu_t: Callable[[float], float] = lambda t: 0, sigma: float = 1.0
+    ):
+        self.mu_t = mu_t
         self.sigma = sigma
 
     def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
@@ -836,11 +1011,11 @@ class BrownianMotion(StochasticProcess):
         assert (dt >= 0).all(), "Increasing time vector required"
         # transposed simulation for automatic broadcasting
         W = rnd.normal(size=(n, t.size))
-        W_drift = W * np.sqrt(dt) * self.sigma + self.mu * dt
+        W_drift = W * np.sqrt(dt) * self.sigma + self.mu_t(t)
         return np.cumsum(W_drift, axis=1)
 
     def expected(self, t: np.ndarray) -> np.ndarray:
-        return self.mu * t
+        return self.mu_t(t)
 
     def stddev(self, t: np.ndarray) -> np.ndarray:
         return np.sqrt(self.sigma ** 2 * t)
@@ -849,8 +1024,10 @@ class BrownianMotion(StochasticProcess):
 class GeometricBrownianMotion(StochasticProcess):
     """Geometric Brownian Motion.(with optional drift)."""
 
-    def __init__(self, mu: float = 0.0, sigma: float = 1.0):
-        self.mu = mu
+    def __init__(
+        self, mu_t: Callable[[float], float] = lambda t: 0, sigma: float = 1.0
+    ):
+        self.mu_t = mu_t
         self.sigma = sigma
 
     def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
@@ -861,13 +1038,13 @@ class GeometricBrownianMotion(StochasticProcess):
         # transposed simulation for automatic broadcasting
         dW = (rnd.normal(size=(t.size, n)).T * np.sqrt(dt)).T
         W = np.cumsum(dW, axis=0)
-        return np.exp(self.sigma * W.T + (self.mu - self.sigma ** 2 / 2) * t)
+        return np.exp(self.sigma * W.T + self.mu_t(t) - self.sigma ** 2 / 2 * t)
 
     def expected(self, t: np.ndarray) -> np.ndarray:
-        return np.exp(self.mu * t)
+        return np.exp(self.mu_t(t))
 
     def stddev(self, t: np.ndarray) -> np.ndarray:
-        return np.sqrt(np.exp(2 * self.mu * t) * (np.exp(self.sigma ** 2 * t) - 1))
+        return np.sqrt(np.exp(2 * self.mu_t(t)) * (np.exp(self.sigma ** 2 * t) - 1))
 
 
 def _get_year_fractions(dategrid: np.ndarray) -> np.ndarray:
@@ -889,14 +1066,106 @@ def simulate_equity_black_scholes_model(
     assert dategrid.dtype == "datetime64[D]"
     ndates = dategrid.size
     yearfractions = _get_year_fractions(dategrid)
-    gbm = GeometricBrownianMotion(r, sigma)
+    gbm = GeometricBrownianMotion(lambda t: r * t, sigma)
     s = S0 * (
         gbm.simulate_with_moment_matching(yearfractions, n, rnd)
         if use_moment_matching
         else gbm.simulate(yearfractions, n, rnd)
     )
     numeraire = np.repeat(np.exp(r * yearfractions).reshape((1, ndates)), n, axis=0)
-    return Model(dategrid, {}, {stock: s}, numeraire, currency)
+    return Model(dategrid, {}, {}, {stock: s}, numeraire, currency)
+
+
+class HoLeeModel(TermStructuresModel):
+
+    h: Final[float] = np.sqrt(np.finfo(float).eps)
+
+    dategrid: Final[np.array]
+    yearfractions: Final[np.array]
+    shortrates: Final[np.ndarray]
+    discount_curve: Optional[
+        Callable[[float], float]
+    ]  # mypy requires hack, see https://ogy.de/mypy-callable-members
+    sigma: Final[float]
+    _mu_t: Final[np.ndarray]
+
+    def __init__(
+        self,
+        dategrid: np.array,
+        discount_curve: Callable[[float], float],
+        sigma: float,
+        n: int,
+        rnd: np.random.RandomState,
+        use_moment_matching: bool = False,
+    ):
+        assert dategrid.dtype == "datetime64[D]"
+        self.dategrid = dategrid
+        self.yearfractions = _get_year_fractions(dategrid)
+        self.discount_curve = discount_curve
+        self.sigma = sigma
+        self._mu_t = self._integral_theta()
+        self.shortrates = self._simulate(n, rnd, use_moment_matching)
+        assert self.shortrates.shape == (n, dategrid.size)
+
+    def _integral_theta(self) -> np.ndarray:
+        assert self.discount_curve
+        dlogBond = (
+            np.log(self.discount_curve(self.yearfractions + self.h))
+            - np.log(self.discount_curve(self.yearfractions))
+        ) / self.h
+        return self.sigma ** 2 * self.yearfractions ** 2 / 2 - dlogBond
+
+    def _integral_mu_t(self, T: float) -> np.ndarray:
+        """Integral over mu_t (i.e. double integral over theta)
+        from model grid to T"""
+        assert T >= self.yearfractions[0]
+        assert T <= self.yearfractions[-1]
+        t = np.linspace(self.yearfractions[0], self.yearfractions[-1], num=100)
+        int_mu = np.cumsum(self.mu_t(t) * np.diff(t, prepend=self.yearfractions[0] - 1))
+        int_mu_fn: Callable[[float], float] = lambda tau: np.interp(tau, t, int_mu)
+        return int_mu_fn(T) - int_mu_fn(self.yearfractions)
+
+    def _simulate(
+        self, n: int, rnd: np.random.RandomState, use_moment_matching: bool
+    ) -> np.ndarray:
+        bm = BrownianMotion(mu_t=self.mu_t, sigma=self.sigma)
+        return (
+            bm.simulate_with_moment_matching(self.yearfractions, n, rnd)
+            if use_moment_matching
+            else bm.simulate(self.yearfractions, n, rnd)
+        )
+
+    def mu_t(self, yearfractions: np.ndarray) -> np.ndarray:
+        return np.interp(yearfractions, self.yearfractions, self._mu_t)
+
+    def linear_rate(self, frequency: str) -> np.ndarray:
+        raise NotImplementedError()
+
+    def discount_factors(self) -> np.ndarray:
+        """Stochastic discount factors implied by the model"""
+        dt = np.diff(self.yearfractions, prepend=0).reshape(1, self.dategrid.size)
+        r_before_0 = np.zeros((self.shortrates.shape[0], 1))
+        r = np.hstack((r_before_0, self.shortrates[:, :-1]))
+        return np.exp(-np.cumsum(r * dt, axis=1))
+
+    def bond_prices(self, T: float) -> np.ndarray:
+        """Stochastic T-bond prices implied by the model,
+        quoted for all times even after T"""
+        t = self.yearfractions
+        lb = (
+            -(T - t) * self.shortrates
+            - self._integral_mu_t(T)
+            + self.sigma ** 2 / 6 * (T - t) ** 3
+        )
+        return np.exp(lb)
+
+    def pathwise_terminal_bond(self) -> np.ndarray:
+        """Pathwise value of the zero coupon bond maturing at model horizon"""
+        dt_reversed = np.diff(self.yearfractions, append=self.yearfractions[-1])[
+            ::-1
+        ].reshape(1, self.dategrid.size)
+        r_reversed = self.shortrates[:, ::-1]
+        return np.exp(-np.cumsum(r_reversed * dt_reversed, axis=1))[:, ::-1]
 
 
 parser = ArgumentParser(description=__doc__)
